"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "utils", {
  enumerable: true,
  get: function get() {
    return _utils["default"];
  }
});
Object.defineProperty(exports, "cardano", {
  enumerable: true,
  get: function get() {
    return _cardano["default"];
  }
});
Object.defineProperty(exports, "CertificateTypes", {
  enumerable: true,
  get: function get() {
    return _cardano.CertificateTypes;
  }
});
Object.defineProperty(exports, "AddressTypeNibbles", {
  enumerable: true,
  get: function get() {
    return _cardano.AddressTypeNibbles;
  }
});
Object.defineProperty(exports, "TxErrors", {
  enumerable: true,
  get: function get() {
    return _cardano.TxErrors;
  }
});
exports["default"] = exports.getErrorDescription = exports.DeviceErrorCodes = exports.TxOutputTypeCodes = void 0;

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _hwTransport = require("@ledgerhq/hw-transport");

var _utils = _interopRequireWildcard(require("./utils"));

var _cardano = _interopRequireWildcard(require("./cardano"));

var _DeviceErrorMessages;

function _createForOfIteratorHelper(o) { if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (o = _unsupportedIterableToArray(o))) { var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var it, normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(n); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var CLA = 0xd7;
var INS = Object.freeze({
  GET_VERSION: 0x00,
  GET_SERIAL: 0x01,
  GET_EXT_PUBLIC_KEY: 0x10,
  DERIVE_ADDRESS: 0x11,
  SIGN_TX: 0x21,
  RUN_TESTS: 0xf0
});
var PoolRegistrationCodes = {
  SIGN_TX_POOL_REGISTRATION_NO: 3,
  SIGN_TX_POOL_REGISTRATION_YES: 4
};
var TxOutputTypeCodes = {
  SIGN_TX_OUTPUT_TYPE_ADDRESS_BYTES: 1,
  SIGN_TX_OUTPUT_TYPE_ADDRESS_PARAMS: 2
};
exports.TxOutputTypeCodes = TxOutputTypeCodes;
var DeviceErrorCodes = {
  ERR_STILL_IN_CALL: 0x6e04,
  // internal
  ERR_INVALID_DATA: 0x6e07,
  ERR_INVALID_BIP_PATH: 0x6e08,
  ERR_REJECTED_BY_USER: 0x6e09,
  ERR_REJECTED_BY_POLICY: 0x6e10,
  ERR_DEVICE_LOCKED: 0x6e11,
  ERR_UNSUPPORTED_ADDRESS_TYPE: 0x6e12,
  // Not thrown by ledger-app-cardano itself but other apps
  ERR_CLA_NOT_SUPPORTED: 0x6e00
};
exports.DeviceErrorCodes = DeviceErrorCodes;
var GH_ERRORS_LINK = "https://github.com/cardano-foundation/ledger-app-cardano/blob/master/src/errors.h";
var DeviceErrorMessages = (_DeviceErrorMessages = {}, (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_INVALID_DATA, "Invalid data supplied to Ledger"), (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_INVALID_BIP_PATH, "Invalid derivation path supplied to Ledger"), (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_REJECTED_BY_USER, "Action rejected by user"), (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_REJECTED_BY_POLICY, "Action rejected by Ledger's security policy"), (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_DEVICE_LOCKED, "Device is locked"), (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_CLA_NOT_SUPPORTED, "Wrong Ledger app"), (0, _defineProperty2["default"])(_DeviceErrorMessages, DeviceErrorCodes.ERR_UNSUPPORTED_ADDRESS_TYPE, "Unsupported address type"), _DeviceErrorMessages);
var Errors = {
  INCORRECT_APP_VERSION: "Operation not supported by the Ledger device, make sure to have the latest version of the Cardano app installed"
};

var getErrorDescription = function getErrorDescription(statusCode) {
  var statusCodeHex = "0x".concat(statusCode.toString(16));
  var defaultMsg = "General error ".concat(statusCodeHex, ". Please consult ").concat(GH_ERRORS_LINK);
  return DeviceErrorMessages[statusCode] || defaultMsg;
}; // It can happen that we try to send a message to the device
// when the device thinks it is still in a middle of previous ADPU stream.
// This happens mostly if host does abort communication for some reason
// leaving ledger mid-call.
// In this case Ledger will respond by ERR_STILL_IN_CALL *and* resetting its state to
// default. We can therefore transparently retry the request.
// Note though that only the *first* request in an multi-APDU exchange should be retried.


exports.getErrorDescription = getErrorDescription;

var wrapRetryStillInCall = function wrapRetryStillInCall(fn) {
  return /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
    var _args = arguments;
    return _regenerator["default"].wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return fn.apply(void 0, _args);

          case 3:
            return _context.abrupt("return", _context.sent);

          case 6:
            _context.prev = 6;
            _context.t0 = _context["catch"](0);

            if (!(_context.t0 && _context.t0.statusCode && _context.t0.statusCode === DeviceErrorCodes.ERR_STILL_IN_CALL)) {
              _context.next = 12;
              break;
            }

            _context.next = 11;
            return fn.apply(void 0, _args);

          case 11:
            return _context.abrupt("return", _context.sent);

          case 12:
            throw _context.t0;

          case 13:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, null, [[0, 6]]);
  }));
};

var wrapConvertError = function wrapConvertError(fn) {
  return /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
    var _args2 = arguments;
    return _regenerator["default"].wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.prev = 0;
            _context2.next = 3;
            return fn.apply(void 0, _args2);

          case 3:
            return _context2.abrupt("return", _context2.sent);

          case 6:
            _context2.prev = 6;
            _context2.t0 = _context2["catch"](0);

            if (_context2.t0 && _context2.t0.statusCode) {
              // keep HwTransport.TransportStatusError
              // just override the message
              _context2.t0.message = "Ledger device: ".concat(getErrorDescription(_context2.t0.statusCode));
            }

            throw _context2.t0;

          case 10:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2, null, [[0, 6]]);
  }));
};
/**
 * Cardano ADA API
 *
 * @example
 * import Ada from "@ledgerhq/hw-app-ada";
 * const ada = new Ada(transport);
 */


var buildSendFn = function buildSendFn(cls, instructionCode) {
  return /*#__PURE__*/function () {
    var _ref3 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(p1, p2, data, expectedResponseLength) {
      var response;
      return _regenerator["default"].wrap(function _callee3$(_context3) {
        while (1) {
          switch (_context3.prev = _context3.next) {
            case 0:
              _context3.next = 2;
              return cls.send(CLA, instructionCode, p1, p2, data);

            case 2:
              response = _context3.sent;
              response = _utils["default"].stripRetcodeFromResponse(response);

              if (expectedResponseLength != null) {
                _utils.Assert.assert(response.length === expectedResponseLength, "unexpected response length: ".concat(response.length, " instead of ").concat(expectedResponseLength));
              }

              return _context3.abrupt("return", response);

            case 6:
            case "end":
              return _context3.stop();
          }
        }
      }, _callee3);
    }));

    return function (_x, _x2, _x3, _x4) {
      return _ref3.apply(this, arguments);
    };
  }();
};

var Ada = /*#__PURE__*/function () {
  function Ada(transport) {
    var scrambleKey = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "ADA";
    (0, _classCallCheck2["default"])(this, Ada);
    this.transport = transport;
    this.methods = ["getVersion", "getSerial", "getExtendedPublicKey", "signTransaction", "deriveAddress", "showAddress"];
    this.transport.decorateAppAPIMethods(this, this.methods, scrambleKey);
    this.send = wrapConvertError(this.transport.send);
  }

  (0, _createClass2["default"])(Ada, [{
    key: "_getVersion",
    value: function () {
      var _getVersion2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
        var _send, P1_UNUSED, P2_UNUSED, response, _response, major, minor, patch, flags_value, FLAG_IS_DEBUG, flags;

        return _regenerator["default"].wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                // moving getVersion() logic to private function in order
                // to disable concurrent execution protection done by this.transport.decorateAppAPIMethods()
                // when invoked from within other calls to check app version
                _send = buildSendFn(this, INS.GET_VERSION);
                P1_UNUSED = 0x00;
                P2_UNUSED = 0x00;
                _context4.next = 5;
                return wrapRetryStillInCall(_send)(P1_UNUSED, P2_UNUSED, _utils["default"].hex_to_buf(""), 4);

              case 5:
                response = _context4.sent;
                _response = (0, _slicedToArray2["default"])(response, 4), major = _response[0], minor = _response[1], patch = _response[2], flags_value = _response[3];
                FLAG_IS_DEBUG = 1; //const FLAG_IS_HEADLESS = 2;

                flags = {
                  isDebug: (flags_value & FLAG_IS_DEBUG) === FLAG_IS_DEBUG
                };
                return _context4.abrupt("return", {
                  major: major,
                  minor: minor,
                  patch: patch,
                  flags: flags
                });

              case 10:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function _getVersion() {
        return _getVersion2.apply(this, arguments);
      }

      return _getVersion;
    }()
    /**
     * Returns an object containing the app version.
     *
     * @returns {Promise<GetVersionResponse>} Result object containing the application version number.
     *
     * @example
     * const { major, minor, patch, flags } = await ada.getVersion();
     * console.log(`App version ${major}.${minor}.${patch}`);
     *
     */

  }, {
    key: "getVersion",
    value: function () {
      var _getVersion3 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5() {
        return _regenerator["default"].wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                return _context5.abrupt("return", this._getVersion());

              case 1:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function getVersion() {
        return _getVersion3.apply(this, arguments);
      }

      return getVersion;
    }()
  }, {
    key: "_isLedgerAppVersionAtLeast",
    value: function () {
      var _isLedgerAppVersionAtLeast2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(minMajor, minMinor) {
        var version, major, minor;
        return _regenerator["default"].wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _context6.next = 2;
                return this._getVersion();

              case 2:
                version = _context6.sent;
                major = parseInt(version.major);
                minor = parseInt(version.minor);

                if (!(isNaN(major) || isNaN(minor))) {
                  _context6.next = 7;
                  break;
                }

                return _context6.abrupt("return", false);

              case 7:
                return _context6.abrupt("return", major >= minMajor && (major > minMajor || minor >= minMinor));

              case 8:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function _isLedgerAppVersionAtLeast(_x5, _x6) {
        return _isLedgerAppVersionAtLeast2.apply(this, arguments);
      }

      return _isLedgerAppVersionAtLeast;
    }()
  }, {
    key: "_ensureLedgerAppVersionAtLeast",
    value: function () {
      var _ensureLedgerAppVersionAtLeast2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7(minMajor, minMinor) {
        var versionCheckSuccess;
        return _regenerator["default"].wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.next = 2;
                return this._isLedgerAppVersionAtLeast(minMajor, minMinor);

              case 2:
                versionCheckSuccess = _context7.sent;

                if (versionCheckSuccess) {
                  _context7.next = 5;
                  break;
                }

                throw new Error(Errors.INCORRECT_APP_VERSION);

              case 5:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _ensureLedgerAppVersionAtLeast(_x7, _x8) {
        return _ensureLedgerAppVersionAtLeast2.apply(this, arguments);
      }

      return _ensureLedgerAppVersionAtLeast;
    }()
    /**
     * Returns an object containing the device serial number.
     *
     * @returns {Promise<GetSerialResponse>} Result object containing the device serial number.
     *
     * @example
     * const { serial } = await ada.getSerial();
     * console.log(`Serial number ${serial}`);
     *
     */

  }, {
    key: "getSerial",
    value: function () {
      var _getSerial = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee8() {
        var _send, P1_UNUSED, P2_UNUSED, response, serial;

        return _regenerator["default"].wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                _context8.next = 2;
                return this._ensureLedgerAppVersionAtLeast(1, 2);

              case 2:
                _send = buildSendFn(this, INS.GET_SERIAL);
                P1_UNUSED = 0x00;
                P2_UNUSED = 0x00;
                _context8.next = 7;
                return wrapRetryStillInCall(_send)(P1_UNUSED, P2_UNUSED, _utils["default"].hex_to_buf(""), 7);

              case 7:
                response = _context8.sent;

                _utils.Assert.assert(response.length === 7);

                serial = _utils["default"].buf_to_hex(response);
                return _context8.abrupt("return", {
                  serial: serial
                });

              case 11:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function getSerial() {
        return _getSerial.apply(this, arguments);
      }

      return getSerial;
    }()
    /**
     * Runs unit tests on the device (DEVEL app build only)
     *
     * @returns {Promise<void>}
     */

  }, {
    key: "runTests",
    value: function () {
      var _runTests = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee9() {
        return _regenerator["default"].wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                _context9.next = 2;
                return wrapRetryStillInCall(this.send)(CLA, INS.RUN_TESTS, 0x00, 0x00);

              case 2:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function runTests() {
        return _runTests.apply(this, arguments);
      }

      return runTests;
    }()
    /**
     * @description Get several public keys; one for each of the specified BIP 32 paths.
     *
     * @param {Array<BIP32Path>} paths The paths. A path must begin with `44'/1815'/account'` or `1852'/1815'/account'`, and may be up to 10 indexes long.
     * @return {Promise<Array<GetExtendedPublicKeyResponse>>} The extended public keys (i.e. with chaincode) for the given paths.
     *
     * @example
     * const [{ publicKey, chainCode }] = await ada.getExtendedPublicKeys([[ HARDENED + 44, HARDENED + 1815, HARDENED + 1 ]]);
     * console.log(publicKey);
     *
     */

  }, {
    key: "getExtendedPublicKeys",
    value: function () {
      var _getExtendedPublicKeys = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee10(paths) {
        var _iterator, _step, path, _send, P1_INIT, P1_NEXT_KEY, P2_UNUSED, result, i, pathData, response, remainingKeysData, _utils$chunkBy, _utils$chunkBy2, publicKey, chainCode, rest;

        return _regenerator["default"].wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                // validate the input
                _utils.Precondition.checkIsArray(paths);

                _iterator = _createForOfIteratorHelper(paths);

                try {
                  for (_iterator.s(); !(_step = _iterator.n()).done;) {
                    path = _step.value;

                    _utils.Precondition.checkIsValidPath(path);
                  }
                } catch (err) {
                  _iterator.e(err);
                } finally {
                  _iterator.f();
                }

                if (!(paths.length > 1)) {
                  _context10.next = 6;
                  break;
                }

                _context10.next = 6;
                return this._ensureLedgerAppVersionAtLeast(2, 1);

              case 6:
                _send = buildSendFn(this, INS.GET_EXT_PUBLIC_KEY);
                P1_INIT = 0x00;
                P1_NEXT_KEY = 0x01;
                P2_UNUSED = 0x00;
                result = [];
                i = 0;

              case 12:
                if (!(i < paths.length)) {
                  _context10.next = 31;
                  break;
                }

                pathData = _cardano["default"].serializeGetExtendedPublicKeyParams(paths[i]);
                response = void 0;

                if (!(i === 0)) {
                  _context10.next = 22;
                  break;
                }

                // initial APDU
                // passing empty Buffer for backwards compatibility
                // of single key export on Ledger app version 2.0.4
                remainingKeysData = paths.length > 1 ? _utils["default"].uint32_to_buf(paths.length - 1) : Buffer.from([]);
                _context10.next = 19;
                return wrapRetryStillInCall(_send)(P1_INIT, P2_UNUSED, Buffer.concat([pathData, remainingKeysData]));

              case 19:
                response = _context10.sent;
                _context10.next = 25;
                break;

              case 22:
                _context10.next = 24;
                return _send(P1_NEXT_KEY, P2_UNUSED, pathData, 64);

              case 24:
                response = _context10.sent;

              case 25:
                _utils$chunkBy = _utils["default"].chunkBy(response, [32, 32]), _utils$chunkBy2 = (0, _slicedToArray2["default"])(_utils$chunkBy, 3), publicKey = _utils$chunkBy2[0], chainCode = _utils$chunkBy2[1], rest = _utils$chunkBy2[2];

                _utils.Assert.assert(rest.length === 0);

                result.push({
                  publicKeyHex: publicKey.toString("hex"),
                  chainCodeHex: chainCode.toString("hex")
                });

              case 28:
                i++;
                _context10.next = 12;
                break;

              case 31:
                return _context10.abrupt("return", result);

              case 32:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function getExtendedPublicKeys(_x9) {
        return _getExtendedPublicKeys.apply(this, arguments);
      }

      return getExtendedPublicKeys;
    }()
    /**
     * @description Get a public key from the specified BIP 32 path.
     *
     * @param {BIP32Path} indexes The path indexes. Path must begin with `44'/1815'/n'`, and may be up to 10 indexes long.
     * @return {Promise<GetExtendedPublicKeyResponse>} The public key with chaincode for the given path.
     *
     * @example
     * const { publicKey, chainCode } = await ada.getExtendedPublicKey([ HARDENED + 44, HARDENED + 1815, HARDENED + 1 ]);
     * console.log(publicKey);
     *
     */

  }, {
    key: "getExtendedPublicKey",
    value: function () {
      var _getExtendedPublicKey = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee11(path) {
        return _regenerator["default"].wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                _context11.next = 2;
                return this.getExtendedPublicKeys([path]);

              case 2:
                return _context11.abrupt("return", _context11.sent[0]);

              case 3:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11, this);
      }));

      function getExtendedPublicKey(_x10) {
        return _getExtendedPublicKey.apply(this, arguments);
      }

      return getExtendedPublicKey;
    }()
    /**
     * @description Gets an address from the specified BIP 32 path.
     *
     * @param {BIP32Path} indexes The path indexes. Path must begin with `(44 or 1852)'/1815'/i'/(0 or 1)/j`, and may be up to 10 indexes long.
     * @return {Promise<DeriveAddressResponse>} The address for the given path.
     *
     * @throws 5001 - The path provided does not have the first 3 indexes hardened or 4th index is not 0, 1 or 2
     * @throws 5002 - The path provided is less than 5 indexes
     * @throws 5003 - Some of the indexes is not a number
     *
     * TODO update error codes
     *
     * @example
     * const { address } = await ada.deriveAddress(
     *   0b1000, // byron address
     *   764824073,
     *   [ HARDENED | 44, HARDENED | 1815, HARDENED | 1, 0, 5 ],
     *   null
     *   null
     *   null
     * );
     *
     * @example
     * const { address } = await ada.deriveAddress(
     *   0b0000, // base address
     *   0x00,
     *   [ HARDENED | 1852, HARDENED | 1815, HARDENED | 0, 0, 5 ],
     *   [ HARDENED | 1852, HARDENED | 1815, HARDENED | 0, 2, 0 ]
     *   null
     *   null
     * );
     *
     */

  }, {
    key: "deriveAddress",
    value: function () {
      var _deriveAddress = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee12(addressTypeNibble, networkIdOrProtocolMagic, spendingPath) {
        var stakingPath,
            stakingKeyHashHex,
            stakingBlockchainPointer,
            _send,
            P1_RETURN,
            P2_UNUSED,
            data,
            response,
            _args12 = arguments;

        return _regenerator["default"].wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                stakingPath = _args12.length > 3 && _args12[3] !== undefined ? _args12[3] : null;
                stakingKeyHashHex = _args12.length > 4 && _args12[4] !== undefined ? _args12[4] : null;
                stakingBlockchainPointer = _args12.length > 5 && _args12[5] !== undefined ? _args12[5] : null;
                _send = buildSendFn(this, INS.DERIVE_ADDRESS);
                P1_RETURN = 0x01;
                P2_UNUSED = 0x00;
                data = _cardano["default"].serializeAddressParams(addressTypeNibble, networkIdOrProtocolMagic, spendingPath, stakingPath, stakingKeyHashHex, stakingBlockchainPointer);
                _context12.next = 9;
                return _send(P1_RETURN, P2_UNUSED, data);

              case 9:
                response = _context12.sent;
                return _context12.abrupt("return", {
                  addressHex: response.toString('hex')
                });

              case 11:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12, this);
      }));

      function deriveAddress(_x11, _x12, _x13) {
        return _deriveAddress.apply(this, arguments);
      }

      return deriveAddress;
    }()
  }, {
    key: "showAddress",
    value: function () {
      var _showAddress = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee13(addressTypeNibble, networkIdOrProtocolMagic, spendingPath) {
        var stakingPath,
            stakingKeyHashHex,
            stakingBlockchainPointer,
            _send,
            P1_DISPLAY,
            P2_UNUSED,
            data,
            _args13 = arguments;

        return _regenerator["default"].wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                stakingPath = _args13.length > 3 && _args13[3] !== undefined ? _args13[3] : null;
                stakingKeyHashHex = _args13.length > 4 && _args13[4] !== undefined ? _args13[4] : null;
                stakingBlockchainPointer = _args13.length > 5 && _args13[5] !== undefined ? _args13[5] : null;
                _send = buildSendFn(this, INS.DERIVE_ADDRESS);
                P1_DISPLAY = 0x02;
                P2_UNUSED = 0x00;
                data = _cardano["default"].serializeAddressParams(addressTypeNibble, networkIdOrProtocolMagic, spendingPath, stakingPath, stakingKeyHashHex, stakingBlockchainPointer);
                _context13.next = 9;
                return _send(P1_DISPLAY, P2_UNUSED, data, 0);

              case 9:
              case "end":
                return _context13.stop();
            }
          }
        }, _callee13, this);
      }));

      function showAddress(_x14, _x15, _x16) {
        return _showAddress.apply(this, arguments);
      }

      return showAddress;
    }()
  }, {
    key: "signTransaction",
    value: function () {
      var _signTransaction = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee25(networkId, protocolMagic, inputs, outputs, feeStr, ttlStr, certificates, withdrawals, metadataHashHex, validityIntervalStartStr) {
        var isSigningPoolRegistrationAsOwner, appHasStakePoolOwnerSupport, appHasMultiassetSupport, containsMultiassets, P1_STAGE_INIT, P1_STAGE_INPUTS, P1_STAGE_OUTPUTS, P1_STAGE_FEE, P1_STAGE_TTL, P1_STAGE_CERTIFICATES, P1_STAGE_WITHDRAWALS, P1_STAGE_METADATA, P1_STAGE_VALIDITY_INTERVAL_START, P1_STAGE_CONFIRM, P1_STAGE_WITNESSES, P2_UNUSED, _send, signTx_init, signTx_addInput, signTx_addOutput, signTx_addCertificate, signTx_addWithdrawal, signTx_setFee, signTx_setTtl, signTx_setMetadata, signTx_setValidityIntervalStart, signTx_awaitConfirm, signTx_getWitness, witnessPaths, owners, witnessOwner, witnessPathsSet, _i, _arr, path, pathKey, _iterator6, _step6, input, _iterator7, _step7, output, _iterator8, _step8, certificate, _iterator9, _step9, withdrawal, _yield$signTx_awaitCo, txHashHex, witnesses, _i2, _witnessPaths, _path, witness;

        return _regenerator["default"].wrap(function _callee25$(_context25) {
          while (1) {
            switch (_context25.prev = _context25.next) {
              case 0:
                // pool registrations are quite restricted
                // this affects witness set construction and many validations
                isSigningPoolRegistrationAsOwner = certificates.some(function (cert) {
                  return cert.type === _cardano.CertificateTypes.STAKE_POOL_REGISTRATION;
                });
                _context25.next = 3;
                return this._isLedgerAppVersionAtLeast(2, 1);

              case 3:
                appHasStakePoolOwnerSupport = _context25.sent;

                if (!(isSigningPoolRegistrationAsOwner && !appHasStakePoolOwnerSupport)) {
                  _context25.next = 6;
                  break;
                }

                throw Error(Errors.INCORRECT_APP_VERSION);

              case 6:
                _context25.next = 8;
                return this._isLedgerAppVersionAtLeast(2, 2);

              case 8:
                appHasMultiassetSupport = _context25.sent;

                if (appHasMultiassetSupport) {
                  _context25.next = 13;
                  break;
                }

                containsMultiassets = outputs.some(function (output) {
                  return output.tokenBundle != null;
                }); // for older app versions:
                // multiasset outputs must not be given
                // validity interval start must not be given
                // ttl must be given

                if (!(containsMultiassets || validityIntervalStartStr != null || ttlStr == null)) {
                  _context25.next = 13;
                  break;
                }

                throw Error(Errors.INCORRECT_APP_VERSION);

              case 13:
                _cardano["default"].validateTransaction(networkId, protocolMagic, inputs, outputs, feeStr, ttlStr, certificates, withdrawals, metadataHashHex, validityIntervalStartStr);

                P1_STAGE_INIT = 0x01;
                P1_STAGE_INPUTS = 0x02;
                P1_STAGE_OUTPUTS = 0x03;
                P1_STAGE_FEE = 0x04;
                P1_STAGE_TTL = 0x05;
                P1_STAGE_CERTIFICATES = 0x06;
                P1_STAGE_WITHDRAWALS = 0x07;
                P1_STAGE_METADATA = 0x08;
                P1_STAGE_VALIDITY_INTERVAL_START = 0x09;
                P1_STAGE_CONFIRM = 0x0a;
                P1_STAGE_WITNESSES = 0x0f;
                P2_UNUSED = 0x00;
                _send = buildSendFn(this, INS.SIGN_TX);

                signTx_init = /*#__PURE__*/function () {
                  var _ref4 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee14(networkId, protocolMagic, numInputs, numOutputs, includeTtl, numCertificates, numWithdrawals, includeMetadata, includeValidityIntervalStart, numWitnesses) {
                    var _serializePoolRegistrationCode, _serializeIncludeInTxData, data, response;

                    return _regenerator["default"].wrap(function _callee14$(_context14) {
                      while (1) {
                        switch (_context14.prev = _context14.next) {
                          case 0:
                            _serializePoolRegistrationCode = function _serializePoolRegistrationCode(isSigningPoolRegistrationAsOwner) {
                              // backwards compatible way of serializing the flag to signalize pool registration
                              // transactions.
                              // TODO To be removed/refactored once Ledger app 2.1 or later is rolled out
                              if (!appHasStakePoolOwnerSupport) {
                                return Buffer.from([]);
                              }

                              return _utils["default"].uint8_to_buf(isSigningPoolRegistrationAsOwner ? PoolRegistrationCodes.SIGN_TX_POOL_REGISTRATION_YES : PoolRegistrationCodes.SIGN_TX_POOL_REGISTRATION_NO);
                            };

                            _serializeIncludeInTxData = function _serializeIncludeInTxData(hasMultiassetSupport) {
                              // TODO remove after Ledger app 2.1 support is not needed anymore
                              if (!hasMultiassetSupport) {
                                return Buffer.concat([_utils["default"].uint8_to_buf(includeMetadata ? _cardano.SignTxIncluded.SIGN_TX_INCLUDED_YES : _cardano.SignTxIncluded.SIGN_TX_INCLUDED_NO)]);
                              }

                              return Buffer.concat([_utils["default"].uint8_to_buf(includeTtl ? _cardano.SignTxIncluded.SIGN_TX_INCLUDED_YES : _cardano.SignTxIncluded.SIGN_TX_INCLUDED_NO), _utils["default"].uint8_to_buf(includeMetadata ? _cardano.SignTxIncluded.SIGN_TX_INCLUDED_YES : _cardano.SignTxIncluded.SIGN_TX_INCLUDED_NO), _utils["default"].uint8_to_buf(includeValidityIntervalStart ? _cardano.SignTxIncluded.SIGN_TX_INCLUDED_YES : _cardano.SignTxIncluded.SIGN_TX_INCLUDED_NO)]);
                            };

                            data = Buffer.concat([_utils["default"].uint8_to_buf(networkId), _utils["default"].uint32_to_buf(protocolMagic), _serializeIncludeInTxData(appHasMultiassetSupport), _serializePoolRegistrationCode(isSigningPoolRegistrationAsOwner), _utils["default"].uint32_to_buf(numInputs), _utils["default"].uint32_to_buf(numOutputs), _utils["default"].uint32_to_buf(numCertificates), _utils["default"].uint32_to_buf(numWithdrawals), _utils["default"].uint32_to_buf(numWitnesses)]);
                            _context14.next = 5;
                            return wrapRetryStillInCall(_send)(P1_STAGE_INIT, P2_UNUSED, data, 0);

                          case 5:
                            response = _context14.sent;

                          case 6:
                          case "end":
                            return _context14.stop();
                        }
                      }
                    }, _callee14);
                  }));

                  return function signTx_init(_x27, _x28, _x29, _x30, _x31, _x32, _x33, _x34, _x35, _x36) {
                    return _ref4.apply(this, arguments);
                  };
                }();

                signTx_addInput = /*#__PURE__*/function () {
                  var _ref5 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee15(input) {
                    var data;
                    return _regenerator["default"].wrap(function _callee15$(_context15) {
                      while (1) {
                        switch (_context15.prev = _context15.next) {
                          case 0:
                            data = Buffer.concat([_utils["default"].hex_to_buf(input.txHashHex), _utils["default"].uint32_to_buf(input.outputIndex)]);
                            _context15.next = 3;
                            return _send(P1_STAGE_INPUTS, P2_UNUSED, data, 0);

                          case 3:
                          case "end":
                            return _context15.stop();
                        }
                      }
                    }, _callee15);
                  }));

                  return function signTx_addInput(_x37) {
                    return _ref5.apply(this, arguments);
                  };
                }();

                signTx_addOutput = /*#__PURE__*/function () {
                  var _ref6 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee16(output) {
                    var P2_BASIC_DATA, P2_ASSET_GROUP, P2_TOKEN, P2_CONFIRM, outputData, outputP2, _iterator2, _step2, assetGroup, data, _iterator3, _step3, token, _data;

                    return _regenerator["default"].wrap(function _callee16$(_context16) {
                      while (1) {
                        switch (_context16.prev = _context16.next) {
                          case 0:
                            P2_BASIC_DATA = 0x30;
                            P2_ASSET_GROUP = 0x31;
                            P2_TOKEN = 0x32;
                            P2_CONFIRM = 0x33; // TODO remove the Before_2_2 version after ledger app 2.2 is rolled out

                            if (appHasMultiassetSupport) {
                              outputData = _cardano["default"].serializeOutputBasicParams(output, protocolMagic, networkId);
                              outputP2 = P2_BASIC_DATA;
                            } else {
                              outputData = _cardano["default"].serializeOutputBasicParamsBefore_2_2(output, protocolMagic, networkId);
                              outputP2 = P2_UNUSED;
                            }

                            _context16.next = 7;
                            return _send(P1_STAGE_OUTPUTS, outputP2, outputData, 0);

                          case 7:
                            if (!(output.tokenBundle != null)) {
                              _context16.next = 44;
                              break;
                            }

                            _iterator2 = _createForOfIteratorHelper(output.tokenBundle);
                            _context16.prev = 9;

                            _iterator2.s();

                          case 11:
                            if ((_step2 = _iterator2.n()).done) {
                              _context16.next = 36;
                              break;
                            }

                            assetGroup = _step2.value;
                            data = Buffer.concat([_utils["default"].hex_to_buf(assetGroup.policyIdHex), _utils["default"].uint32_to_buf(assetGroup.tokens.length)]);
                            _context16.next = 16;
                            return _send(P1_STAGE_OUTPUTS, P2_ASSET_GROUP, data, 0);

                          case 16:
                            _iterator3 = _createForOfIteratorHelper(assetGroup.tokens);
                            _context16.prev = 17;

                            _iterator3.s();

                          case 19:
                            if ((_step3 = _iterator3.n()).done) {
                              _context16.next = 26;
                              break;
                            }

                            token = _step3.value;
                            _data = Buffer.concat([_utils["default"].uint32_to_buf(token.assetNameHex.length / 2), _utils["default"].hex_to_buf(token.assetNameHex), _utils["default"].uint64_to_buf(token.amountStr)]);
                            _context16.next = 24;
                            return _send(P1_STAGE_OUTPUTS, P2_TOKEN, _data, 0);

                          case 24:
                            _context16.next = 19;
                            break;

                          case 26:
                            _context16.next = 31;
                            break;

                          case 28:
                            _context16.prev = 28;
                            _context16.t0 = _context16["catch"](17);

                            _iterator3.e(_context16.t0);

                          case 31:
                            _context16.prev = 31;

                            _iterator3.f();

                            return _context16.finish(31);

                          case 34:
                            _context16.next = 11;
                            break;

                          case 36:
                            _context16.next = 41;
                            break;

                          case 38:
                            _context16.prev = 38;
                            _context16.t1 = _context16["catch"](9);

                            _iterator2.e(_context16.t1);

                          case 41:
                            _context16.prev = 41;

                            _iterator2.f();

                            return _context16.finish(41);

                          case 44:
                            if (!appHasMultiassetSupport) {
                              _context16.next = 47;
                              break;
                            }

                            _context16.next = 47;
                            return _send(P1_STAGE_OUTPUTS, P2_CONFIRM, Buffer.alloc(0), 0);

                          case 47:
                          case "end":
                            return _context16.stop();
                        }
                      }
                    }, _callee16, null, [[9, 38, 41, 44], [17, 28, 31, 34]]);
                  }));

                  return function signTx_addOutput(_x38) {
                    return _ref6.apply(this, arguments);
                  };
                }();

                signTx_addCertificate = /*#__PURE__*/function () {
                  var _ref7 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee17(type, path, poolKeyHashHex, poolParams) {
                    var dataFields, data, APDU_INSTRUCTIONS, _iterator4, _step4, owner, _iterator5, _step5, relay;

                    return _regenerator["default"].wrap(function _callee17$(_context17) {
                      while (1) {
                        switch (_context17.prev = _context17.next) {
                          case 0:
                            dataFields = [_utils["default"].uint8_to_buf(type)];
                            _context17.t0 = type;
                            _context17.next = _context17.t0 === _cardano.CertificateTypes.STAKE_REGISTRATION ? 4 : _context17.t0 === _cardano.CertificateTypes.STAKE_DEREGISTRATION ? 4 : _context17.t0 === _cardano.CertificateTypes.STAKE_DELEGATION ? 4 : _context17.t0 === _cardano.CertificateTypes.STAKE_POOL_REGISTRATION ? 6 : 8;
                            break;

                          case 4:
                            if (path != null) dataFields.push(_utils["default"].path_to_buf(path));
                            return _context17.abrupt("break", 9);

                          case 6:
                            _utils.Assert.assert(isSigningPoolRegistrationAsOwner, "tx certificates validation messed up"); // OK, data are serialized and sent later in additional APDUs


                            return _context17.abrupt("break", 9);

                          case 8:
                            // validation should catch invalid certificate type
                            _utils.Assert.assert(false, "invalid certificate type");

                          case 9:
                            if (poolKeyHashHex != null) {
                              dataFields.push(_utils["default"].hex_to_buf(poolKeyHashHex));
                            }

                            data = Buffer.concat(dataFields);
                            _context17.next = 13;
                            return _send(P1_STAGE_CERTIFICATES, P2_UNUSED, data, 0);

                          case 13:
                            if (!(type === _cardano.CertificateTypes.STAKE_POOL_REGISTRATION)) {
                              _context17.next = 56;
                              break;
                            }

                            (0, _utils.invariant)(poolParams != null); // additional data for pool certificate

                            APDU_INSTRUCTIONS = {
                              POOL_PARAMS: 0x30,
                              OWNERS: 0x31,
                              RELAYS: 0x32,
                              METADATA: 0x33,
                              CONFIRMATION: 0x34
                            };
                            _context17.next = 18;
                            return _send(P1_STAGE_CERTIFICATES, APDU_INSTRUCTIONS.POOL_PARAMS, _cardano["default"].serializePoolInitialParams(poolParams), 0);

                          case 18:
                            _iterator4 = _createForOfIteratorHelper(poolParams.poolOwners);
                            _context17.prev = 19;

                            _iterator4.s();

                          case 21:
                            if ((_step4 = _iterator4.n()).done) {
                              _context17.next = 27;
                              break;
                            }

                            owner = _step4.value;
                            _context17.next = 25;
                            return _send(P1_STAGE_CERTIFICATES, APDU_INSTRUCTIONS.OWNERS, _cardano["default"].serializePoolOwnerParams(owner), 0);

                          case 25:
                            _context17.next = 21;
                            break;

                          case 27:
                            _context17.next = 32;
                            break;

                          case 29:
                            _context17.prev = 29;
                            _context17.t1 = _context17["catch"](19);

                            _iterator4.e(_context17.t1);

                          case 32:
                            _context17.prev = 32;

                            _iterator4.f();

                            return _context17.finish(32);

                          case 35:
                            _iterator5 = _createForOfIteratorHelper(poolParams.relays);
                            _context17.prev = 36;

                            _iterator5.s();

                          case 38:
                            if ((_step5 = _iterator5.n()).done) {
                              _context17.next = 44;
                              break;
                            }

                            relay = _step5.value;
                            _context17.next = 42;
                            return _send(P1_STAGE_CERTIFICATES, APDU_INSTRUCTIONS.RELAYS, _cardano["default"].serializePoolRelayParams(relay), 0);

                          case 42:
                            _context17.next = 38;
                            break;

                          case 44:
                            _context17.next = 49;
                            break;

                          case 46:
                            _context17.prev = 46;
                            _context17.t2 = _context17["catch"](36);

                            _iterator5.e(_context17.t2);

                          case 49:
                            _context17.prev = 49;

                            _iterator5.f();

                            return _context17.finish(49);

                          case 52:
                            _context17.next = 54;
                            return _send(P1_STAGE_CERTIFICATES, APDU_INSTRUCTIONS.METADATA, _cardano["default"].serializePoolMetadataParams(poolParams.metadata), 0);

                          case 54:
                            _context17.next = 56;
                            return _send(P1_STAGE_CERTIFICATES, APDU_INSTRUCTIONS.CONFIRMATION, Buffer.alloc(0), 0);

                          case 56:
                          case "end":
                            return _context17.stop();
                        }
                      }
                    }, _callee17, null, [[19, 29, 32, 35], [36, 46, 49, 52]]);
                  }));

                  return function signTx_addCertificate(_x39, _x40, _x41, _x42) {
                    return _ref7.apply(this, arguments);
                  };
                }();

                signTx_addWithdrawal = /*#__PURE__*/function () {
                  var _ref8 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee18(path, amountStr) {
                    var data;
                    return _regenerator["default"].wrap(function _callee18$(_context18) {
                      while (1) {
                        switch (_context18.prev = _context18.next) {
                          case 0:
                            data = Buffer.concat([_utils["default"].ada_amount_to_buf(amountStr), _utils["default"].path_to_buf(path)]);
                            _context18.next = 3;
                            return _send(P1_STAGE_WITHDRAWALS, P2_UNUSED, data, 0);

                          case 3:
                          case "end":
                            return _context18.stop();
                        }
                      }
                    }, _callee18);
                  }));

                  return function signTx_addWithdrawal(_x43, _x44) {
                    return _ref8.apply(this, arguments);
                  };
                }();

                signTx_setFee = /*#__PURE__*/function () {
                  var _ref9 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee19(feeStr) {
                    var data;
                    return _regenerator["default"].wrap(function _callee19$(_context19) {
                      while (1) {
                        switch (_context19.prev = _context19.next) {
                          case 0:
                            data = Buffer.concat([_utils["default"].ada_amount_to_buf(feeStr)]);
                            _context19.next = 3;
                            return _send(P1_STAGE_FEE, P2_UNUSED, data, 0);

                          case 3:
                          case "end":
                            return _context19.stop();
                        }
                      }
                    }, _callee19);
                  }));

                  return function signTx_setFee(_x45) {
                    return _ref9.apply(this, arguments);
                  };
                }();

                signTx_setTtl = /*#__PURE__*/function () {
                  var _ref10 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee20(ttlStr) {
                    var data;
                    return _regenerator["default"].wrap(function _callee20$(_context20) {
                      while (1) {
                        switch (_context20.prev = _context20.next) {
                          case 0:
                            data = Buffer.concat([_utils["default"].uint64_to_buf(ttlStr)]);
                            _context20.next = 3;
                            return _send(P1_STAGE_TTL, P2_UNUSED, data, 0);

                          case 3:
                          case "end":
                            return _context20.stop();
                        }
                      }
                    }, _callee20);
                  }));

                  return function signTx_setTtl(_x46) {
                    return _ref10.apply(this, arguments);
                  };
                }();

                signTx_setMetadata = /*#__PURE__*/function () {
                  var _ref11 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee21(metadataHashHex) {
                    var data;
                    return _regenerator["default"].wrap(function _callee21$(_context21) {
                      while (1) {
                        switch (_context21.prev = _context21.next) {
                          case 0:
                            data = _utils["default"].hex_to_buf(metadataHashHex);
                            _context21.next = 3;
                            return _send(P1_STAGE_METADATA, P2_UNUSED, data, 0);

                          case 3:
                          case "end":
                            return _context21.stop();
                        }
                      }
                    }, _callee21);
                  }));

                  return function signTx_setMetadata(_x47) {
                    return _ref11.apply(this, arguments);
                  };
                }();

                signTx_setValidityIntervalStart = /*#__PURE__*/function () {
                  var _ref12 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee22(validityIntervalStartStr) {
                    var data;
                    return _regenerator["default"].wrap(function _callee22$(_context22) {
                      while (1) {
                        switch (_context22.prev = _context22.next) {
                          case 0:
                            data = Buffer.concat([_utils["default"].uint64_to_buf(validityIntervalStartStr)]);
                            _context22.next = 3;
                            return _send(P1_STAGE_VALIDITY_INTERVAL_START, P2_UNUSED, data);

                          case 3:
                          case "end":
                            return _context22.stop();
                        }
                      }
                    }, _callee22);
                  }));

                  return function signTx_setValidityIntervalStart(_x48) {
                    return _ref12.apply(this, arguments);
                  };
                }();

                signTx_awaitConfirm = /*#__PURE__*/function () {
                  var _ref13 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee23() {
                    var confirmP1, response;
                    return _regenerator["default"].wrap(function _callee23$(_context23) {
                      while (1) {
                        switch (_context23.prev = _context23.next) {
                          case 0:
                            // TODO remove after ledger app 2.2 is rolled out
                            if (appHasMultiassetSupport) {
                              confirmP1 = P1_STAGE_CONFIRM;
                            } else {
                              confirmP1 = 0x09; // needed for backward compatibility
                            }

                            _context23.next = 3;
                            return _send(confirmP1, P2_UNUSED, _utils["default"].hex_to_buf(""), _cardano["default"].TX_HASH_LENGTH);

                          case 3:
                            response = _context23.sent;
                            return _context23.abrupt("return", {
                              txHashHex: response.toString("hex")
                            });

                          case 5:
                          case "end":
                            return _context23.stop();
                        }
                      }
                    }, _callee23);
                  }));

                  return function signTx_awaitConfirm() {
                    return _ref13.apply(this, arguments);
                  };
                }();

                signTx_getWitness = /*#__PURE__*/function () {
                  var _ref14 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee24(path) {
                    var witnessP1, data, response;
                    return _regenerator["default"].wrap(function _callee24$(_context24) {
                      while (1) {
                        switch (_context24.prev = _context24.next) {
                          case 0:
                            // TODO remove after ledger app 2.2 is rolled out
                            if (appHasMultiassetSupport) {
                              witnessP1 = P1_STAGE_WITNESSES;
                            } else {
                              witnessP1 = 0x0a; // needed for backward compatibility
                            }

                            data = Buffer.concat([_utils["default"].path_to_buf(path)]);
                            _context24.next = 4;
                            return _send(witnessP1, P2_UNUSED, data, 64);

                          case 4:
                            response = _context24.sent;
                            return _context24.abrupt("return", {
                              path: path,
                              witnessSignatureHex: _utils["default"].buf_to_hex(response)
                            });

                          case 6:
                          case "end":
                            return _context24.stop();
                        }
                      }
                    }, _callee24);
                  }));

                  return function signTx_getWitness(_x49) {
                    return _ref14.apply(this, arguments);
                  };
                }();

                witnessPaths = [];

                if (isSigningPoolRegistrationAsOwner) {
                  // there should be exactly one owner given by path which will be used for the witness
                  _utils.Assert.assert(certificates.length == 1);

                  (0, _utils.invariant)(certificates[0].poolRegistrationParams != null);
                  owners = certificates[0].poolRegistrationParams.poolOwners;
                  witnessOwner = owners.find(function (owner) {
                    return !!owner.stakingPath;
                  });
                  (0, _utils.invariant)(witnessOwner != null);
                  witnessPaths.push(witnessOwner.stakingPath);
                } else {
                  // we collect required witnesses for inputs, certificates and withdrawals
                  // each path is included only once
                  witnessPathsSet = new Set();

                  for (_i = 0, _arr = [].concat((0, _toConsumableArray2["default"])(inputs), (0, _toConsumableArray2["default"])(certificates), (0, _toConsumableArray2["default"])(withdrawals)); _i < _arr.length; _i++) {
                    path = _arr[_i].path;
                    pathKey = JSON.stringify(path);

                    if (!witnessPathsSet.has(pathKey)) {
                      witnessPathsSet.add(pathKey);
                      witnessPaths.push(path);
                    }
                  }
                }

                _context25.next = 42;
                return signTx_init(networkId, protocolMagic, inputs.length, outputs.length, ttlStr != null, certificates.length, withdrawals.length, metadataHashHex != null, validityIntervalStartStr != null, witnessPaths.length);

              case 42:
                // inputs
                _iterator6 = _createForOfIteratorHelper(inputs);
                _context25.prev = 43;

                _iterator6.s();

              case 45:
                if ((_step6 = _iterator6.n()).done) {
                  _context25.next = 51;
                  break;
                }

                input = _step6.value;
                _context25.next = 49;
                return signTx_addInput(input);

              case 49:
                _context25.next = 45;
                break;

              case 51:
                _context25.next = 56;
                break;

              case 53:
                _context25.prev = 53;
                _context25.t0 = _context25["catch"](43);

                _iterator6.e(_context25.t0);

              case 56:
                _context25.prev = 56;

                _iterator6.f();

                return _context25.finish(56);

              case 59:
                // outputs
                _iterator7 = _createForOfIteratorHelper(outputs);
                _context25.prev = 60;

                _iterator7.s();

              case 62:
                if ((_step7 = _iterator7.n()).done) {
                  _context25.next = 68;
                  break;
                }

                output = _step7.value;
                _context25.next = 66;
                return signTx_addOutput(output);

              case 66:
                _context25.next = 62;
                break;

              case 68:
                _context25.next = 73;
                break;

              case 70:
                _context25.prev = 70;
                _context25.t1 = _context25["catch"](60);

                _iterator7.e(_context25.t1);

              case 73:
                _context25.prev = 73;

                _iterator7.f();

                return _context25.finish(73);

              case 76:
                _context25.next = 78;
                return signTx_setFee(feeStr);

              case 78:
                if (!(ttlStr != null)) {
                  _context25.next = 81;
                  break;
                }

                _context25.next = 81;
                return signTx_setTtl(ttlStr);

              case 81:
                if (!(certificates.length > 0)) {
                  _context25.next = 99;
                  break;
                }

                _iterator8 = _createForOfIteratorHelper(certificates);
                _context25.prev = 83;

                _iterator8.s();

              case 85:
                if ((_step8 = _iterator8.n()).done) {
                  _context25.next = 91;
                  break;
                }

                certificate = _step8.value;
                _context25.next = 89;
                return signTx_addCertificate(certificate.type, certificate.path, certificate.poolKeyHashHex, certificate.poolRegistrationParams);

              case 89:
                _context25.next = 85;
                break;

              case 91:
                _context25.next = 96;
                break;

              case 93:
                _context25.prev = 93;
                _context25.t2 = _context25["catch"](83);

                _iterator8.e(_context25.t2);

              case 96:
                _context25.prev = 96;

                _iterator8.f();

                return _context25.finish(96);

              case 99:
                if (!(withdrawals.length > 0)) {
                  _context25.next = 117;
                  break;
                }

                _iterator9 = _createForOfIteratorHelper(withdrawals);
                _context25.prev = 101;

                _iterator9.s();

              case 103:
                if ((_step9 = _iterator9.n()).done) {
                  _context25.next = 109;
                  break;
                }

                withdrawal = _step9.value;
                _context25.next = 107;
                return signTx_addWithdrawal(withdrawal.path, withdrawal.amountStr);

              case 107:
                _context25.next = 103;
                break;

              case 109:
                _context25.next = 114;
                break;

              case 111:
                _context25.prev = 111;
                _context25.t3 = _context25["catch"](101);

                _iterator9.e(_context25.t3);

              case 114:
                _context25.prev = 114;

                _iterator9.f();

                return _context25.finish(114);

              case 117:
                if (!(metadataHashHex != null)) {
                  _context25.next = 120;
                  break;
                }

                _context25.next = 120;
                return signTx_setMetadata(metadataHashHex);

              case 120:
                if (!(validityIntervalStartStr != null)) {
                  _context25.next = 123;
                  break;
                }

                _context25.next = 123;
                return signTx_setValidityIntervalStart(validityIntervalStartStr);

              case 123:
                _context25.next = 125;
                return signTx_awaitConfirm();

              case 125:
                _yield$signTx_awaitCo = _context25.sent;
                txHashHex = _yield$signTx_awaitCo.txHashHex;
                // witnesses
                witnesses = [];
                _i2 = 0, _witnessPaths = witnessPaths;

              case 129:
                if (!(_i2 < _witnessPaths.length)) {
                  _context25.next = 140;
                  break;
                }

                _path = _witnessPaths[_i2];
                (0, _utils.invariant)(_path != null);

                _utils.Precondition.checkIsValidPath(_path, "Invalid path to witness has been supplied");

                _context25.next = 135;
                return signTx_getWitness(_path);

              case 135:
                witness = _context25.sent;
                witnesses.push(witness);

              case 137:
                _i2++;
                _context25.next = 129;
                break;

              case 140:
                return _context25.abrupt("return", {
                  txHashHex: txHashHex,
                  witnesses: witnesses
                });

              case 141:
              case "end":
                return _context25.stop();
            }
          }
        }, _callee25, this, [[43, 53, 56, 59], [60, 70, 73, 76], [83, 93, 96, 99], [101, 111, 114, 117]]);
      }));

      function signTransaction(_x17, _x18, _x19, _x20, _x21, _x22, _x23, _x24, _x25, _x26) {
        return _signTransaction.apply(this, arguments);
      }

      return signTransaction;
    }()
  }]);
  return Ada;
}(); // reexport


exports["default"] = Ada;
//# sourceMappingURL=Ada.js.map