"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.validateTransaction = validateTransaction;
exports.serializeAddressParams = serializeAddressParams;
exports.serializeOutputBasicParams = serializeOutputBasicParams;
exports.serializeOutputBasicParamsBefore_2_2 = serializeOutputBasicParamsBefore_2_2;
exports.serializePoolInitialParams = serializePoolInitialParams;
exports.serializePoolOwnerParams = serializePoolOwnerParams;
exports.serializePoolRelayParams = serializePoolRelayParams;
exports.serializePoolMetadataParams = serializePoolMetadataParams;
exports.serializeGetExtendedPublicKeyParams = serializeGetExtendedPublicKeyParams;
exports["default"] = exports.TxErrors = exports.GetKeyErrors = exports.SignTxIncluded = exports.CertificateTypes = exports.AddressTypeNibbles = void 0;

var _utils = _interopRequireWildcard(require("./utils"));

var _Ada = require("./Ada");

var _utils2 = require("../lib/utils");

function _createForOfIteratorHelper(o) { if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (o = _unsupportedIterableToArray(o))) { var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var it, normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(n); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var HARDENED = 0x80000000;
var AddressTypeNibbles = Object.freeze({
  BASE: 0,
  POINTER: 4,
  ENTERPRISE: 6,
  BYRON: 8,
  REWARD: 14
});
exports.AddressTypeNibbles = AddressTypeNibbles;
var CertificateTypes = Object.freeze({
  STAKE_REGISTRATION: 0,
  STAKE_DEREGISTRATION: 1,
  STAKE_DELEGATION: 2,
  STAKE_POOL_REGISTRATION: 3
});
exports.CertificateTypes = CertificateTypes;
var SignTxIncluded = Object.freeze({
  SIGN_TX_INCLUDED_NO: 1,
  SIGN_TX_INCLUDED_YES: 2
});
exports.SignTxIncluded = SignTxIncluded;
var KEY_HASH_LENGTH = 28;
var TX_HASH_LENGTH = 32;
var TOKEN_POLICY_LENGTH = 28;
var TOKEN_NAME_LENGTH = 32;
var ASSET_GROUPS_MAX = 1000;
var TOKENS_IN_GROUP_MAX = 1000;
var POOL_REGISTRATION_OWNERS_MAX = 1000;
var POOL_REGISTRATION_RELAYS_MAX = 1000;
var GetKeyErrors = {
  INVALID_PATH: "invalid key path"
};
exports.GetKeyErrors = GetKeyErrors;
var TxErrors = {
  INVALID_PROTOCOL_MAGIC: "invalid protocol magic",
  INVALID_NETWORK_ID: "invalid network id",
  INPUTS_NOT_ARRAY: "inputs not an array",
  INPUT_WITH_PATH: "stake pool registration: inputs should not contain the witness path",
  INPUT_INVALID_TX_HASH: "invalid tx hash in an input",
  OUTPUTS_NOT_ARRAY: "outputs not an array",
  OUTPUT_INVALID_AMOUNT: "invalid amount in an output",
  OUTPUT_INVALID_TOKEN_BUNDLE: "invalid multiasset token bundle in an output",
  OUTPUT_INVALID_TOKEN_POLICY: "invalid multiasset token policy",
  OUTPUT_INVALID_ASSET_NAME: "invalid asset name in the token bundle in an output",
  OUTPUT_INVALID_ADDRESS: "invalid address in an output",
  OUTPUT_WITH_PATH: "outputs given by path are not allowed for stake pool registration transactions",
  OUTPUT_UNKNOWN_TYPE: "unknown output type",
  OUTPUT_INVALID_SPENDING_PATH: "invalid spending path in an output",
  OUTPUT_INVALID_BLOCKCHAIN_POINTER: "invalid blockchain pointer in an output",
  OUTPUT_INVALID_STAKING_KEY_PATH: "invalid staking key path in an output",
  OUTPUT_INVALID_STAKING_KEY_HASH: "invalid staking key hash in an output",
  OUTPUT_INVALID_STAKING_INFO: "Invalid staking info in an output",
  FEE_INVALID: "invalid fee",
  TTL_INVALID: "invalid ttl",
  CERTIFICATES_NOT_ARRAY: "certificates not an array",
  CERTIFICATES_COMBINATION_FORBIDDEN: "pool registration must not be combined with other certificates",
  CERTIFICATE_MISSING_PATH: "path is required for one of the certificates",
  CERTIFICATE_SUPERFLUOUS_PATH: "superfluous path in a certificate",
  CERTIFICATE_MISSING_POOL_KEY_HASH: "pool key hash missing in a certificate",
  CERTIFICATE_SUPERFLUOUS_POOL_KEY_HASH: "superfluous pool key hash in a certificate",
  CERTIFICATE_INVALID_POOL_KEY_HASH: "invalid pool key hash in a certificate",
  CERTIFICATE_INVALID: "invalid certificate",
  CERTIFICATE_POOL_MISSING_POOL_PARAMS: "missing stake pool params in a pool registration certificate",
  CERTIFICATE_POOL_INVALID_POOL_KEY_HASH: "invalid pool key hash in a pool registration certificate",
  CERTIFICATE_POOL_INVALID_VRF_KEY_HASH: "invalid vrf key hash in a pool registration certificate",
  CERTIFICATE_POOL_INVALID_PLEDGE: "invalid pledge in a pool registration certificate",
  CERTIFICATE_POOL_INVALID_COST: "invalid cost in a pool registration certificate",
  CERTIFICATE_POOL_INVALID_MARGIN: "invalid margin in a pool registration certificate",
  CERTIFICATE_POOL_INVALID_MARGIN_DENOMINATOR: "pool margin denominator must be a value between 1 and 10^15",
  CERTIFICATE_POOL_INVALID_REWARD_ACCOUNT: "invalid reward account in a pool registration certificate",
  CERTIFICATE_POOL_OWNERS_NOT_ARRAY: "owners not an array in a pool registration certificate",
  CERTIFICATE_POOL_OWNERS_TOO_MANY: "too many owners in a pool registration certificate",
  CERTIFICATE_POOL_OWNERS_SINGLE_PATH: "there should be exactly one owner given by path in a pool registration certificate",
  CERTIFICATE_POOL_OWNER_INCOMPLETE: "incomplete owner params in a pool registration certificate",
  CERTIFICATE_POOL_OWNER_INVALID_PATH: "invalid owner path in a pool registration certificate",
  CERTIFICATE_POOL_OWNER_INVALID_KEY_HASH: "invalid owner key hash in a pool registration certificate",
  CERTIFICATE_POOL_RELAYS_NOT_ARRAY: "relays not an array in a pool registration certificate",
  CERTIFICATE_POOL_RELAYS_TOO_MANY: "too many pool relays in a pool registration certificate",
  CERTIFICATE_POOL_RELAY_INVALID_TYPE: "invalid type of a relay in a pool registration certificate",
  CERTIFICATE_POOL_RELAY_INVALID_PORT: "invalid port in a relay in a pool registration certificate",
  CERTIFICATE_POOL_RELAY_INVALID_IPV4: "invalid ipv4 in a relay in a pool registration certificate",
  CERTIFICATE_POOL_RELAY_INVALID_IPV6: "invalid ipv6 in a relay in a pool registration certificate",
  CERTIFICATE_POOL_RELAY_INVALID_DNS: "invalid dns record in a relay in a pool registration certificate",
  CERTIFICATE_POOL_RELAY_MISSING_DNS: "missing dns record in a relay in a pool registration certificate",
  CERTIFICATE_POOL_METADATA_INVALID_URL: "invalid metadata in a pool registration certificate: invalid url",
  CERTIFICATE_POOL_METADATA_INVALID_HASH: "invalid metadata in a pool registration certificate: invalid hash",
  WITHDRAWALS_NOT_ARRAY: "withdrawals not an array",
  WITHDRAWALS_FORBIDDEN: "no withdrawals allowed for transactions registering stake pools",
  METADATA_INVALID: "invalid metadata",
  VALIDITY_INTERVAL_START_INVALID: "invalid validity interval start"
};
exports.TxErrors = TxErrors;

function validateCertificates(certificates) {
  _utils.Precondition.checkIsArray(certificates, TxErrors.CERTIFICATES_NOT_ARRAY);

  var isSigningPoolRegistrationAsOwner = certificates.some(function (cert) {
    return cert.type === CertificateTypes.STAKE_POOL_REGISTRATION;
  });

  var _iterator = _createForOfIteratorHelper(certificates),
      _step;

  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var cert = _step.value;
      if (!cert) throw new Error(TxErrors.CERTIFICATE_INVALID);

      switch (cert.type) {
        case CertificateTypes.STAKE_REGISTRATION:
          {
            _utils.Precondition.check(!isSigningPoolRegistrationAsOwner, TxErrors.CERTIFICATES_COMBINATION_FORBIDDEN);

            _utils.Precondition.checkIsValidPath(cert.path, TxErrors.CERTIFICATE_MISSING_PATH);

            _utils.Precondition.check(!cert.poolKeyHashHex, TxErrors.CERTIFICATE_SUPERFLUOUS_POOL_KEY_HASH);

            break;
          }

        case CertificateTypes.STAKE_DEREGISTRATION:
          {
            _utils.Precondition.check(!isSigningPoolRegistrationAsOwner, TxErrors.CERTIFICATES_COMBINATION_FORBIDDEN);

            _utils.Precondition.checkIsValidPath(cert.path, TxErrors.CERTIFICATE_MISSING_PATH);

            _utils.Precondition.check(!cert.poolKeyHashHex, TxErrors.CERTIFICATE_SUPERFLUOUS_POOL_KEY_HASH);

            break;
          }

        case CertificateTypes.STAKE_DELEGATION:
          {
            _utils.Precondition.check(!isSigningPoolRegistrationAsOwner, TxErrors.CERTIFICATES_COMBINATION_FORBIDDEN);

            _utils.Precondition.checkIsValidPath(cert.path, TxErrors.CERTIFICATE_MISSING_PATH);

            _utils.Precondition.checkIsHexString(cert.poolKeyHashHex, TxErrors.CERTIFICATE_MISSING_POOL_KEY_HASH);

            _utils.Precondition.check(cert.poolKeyHashHex.length == KEY_HASH_LENGTH * 2, TxErrors.CERTIFICATE_INVALID_POOL_KEY_HASH);

            break;
          }

        case CertificateTypes.STAKE_POOL_REGISTRATION:
          {
            _utils.Assert.assert(isSigningPoolRegistrationAsOwner);

            _utils.Precondition.check(!cert.path, TxErrors.CERTIFICATE_SUPERFLUOUS_PATH);

            var poolParams = cert.poolRegistrationParams;

            _utils.Precondition.check(!!poolParams, TxErrors.CERTIFICATE_POOL_MISSING_POOL_PARAMS); // serialization succeeds if and only if params are valid


            serializePoolInitialParams(poolParams); // owners

            _utils.Precondition.checkIsArray(poolParams.poolOwners, TxErrors.CERTIFICATE_POOL_OWNERS_NOT_ARRAY);

            var numPathOwners = 0;

            var _iterator2 = _createForOfIteratorHelper(poolParams.poolOwners),
                _step2;

            try {
              for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
                var owner = _step2.value;

                if (owner.stakingPath) {
                  numPathOwners++;
                  serializePoolOwnerParams(owner);
                }
              }
            } catch (err) {
              _iterator2.e(err);
            } finally {
              _iterator2.f();
            }

            if (numPathOwners !== 1) throw new Error(TxErrors.CERTIFICATE_POOL_OWNERS_SINGLE_PATH); // relays

            _utils.Precondition.checkIsArray(poolParams.relays, TxErrors.CERTIFICATE_POOL_RELAYS_NOT_ARRAY);

            var _iterator3 = _createForOfIteratorHelper(poolParams.relays),
                _step3;

            try {
              for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
                var relay = _step3.value;
                serializePoolRelayParams(relay);
              } // metadata

            } catch (err) {
              _iterator3.e(err);
            } finally {
              _iterator3.f();
            }

            serializePoolMetadataParams(poolParams.metadata);
            break;
          }

        default:
          throw new Error(TxErrors.CERTIFICATE_INVALID);
      }
    }
  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }
}

function validateTransaction(networkId, protocolMagic, inputs, outputs, feeStr, ttlStr, certificates, withdrawals, metadataHashHex, validityIntervalStartStr) {
  _utils.Precondition.checkIsArray(certificates, TxErrors.CERTIFICATES_NOT_ARRAY);

  var isSigningPoolRegistrationAsOwner = certificates.some(function (cert) {
    return cert.type === CertificateTypes.STAKE_POOL_REGISTRATION;
  }); // inputs

  _utils.Precondition.checkIsArray(inputs, TxErrors.INPUTS_NOT_ARRAY);

  var _iterator4 = _createForOfIteratorHelper(inputs),
      _step4;

  try {
    for (_iterator4.s(); !(_step4 = _iterator4.n()).done;) {
      var input = _step4.value;

      _utils.Precondition.checkIsHexString(input.txHashHex, TxErrors.INPUT_INVALID_TX_HASH);

      _utils.Precondition.check(input.txHashHex.length === TX_HASH_LENGTH * 2, TxErrors.INPUT_INVALID_TX_HASH);

      if (isSigningPoolRegistrationAsOwner) {
        // input should not be given with a path
        // the path is not used, but we check just to avoid potential confusion of developers using this
        _utils.Precondition.check(!input.path, TxErrors.INPUT_WITH_PATH);
      }
    } // outputs

  } catch (err) {
    _iterator4.e(err);
  } finally {
    _iterator4.f();
  }

  _utils.Precondition.checkIsArray(outputs, TxErrors.OUTPUTS_NOT_ARRAY);

  var _iterator5 = _createForOfIteratorHelper(outputs),
      _step5;

  try {
    for (_iterator5.s(); !(_step5 = _iterator5.n()).done;) {
      var output = _step5.value;
      // we try to serialize the data, an error is thrown if ada amount or address params are invalid
      serializeOutputBasicParams(output, protocolMagic, networkId);

      if (output.spendingPath) {
        _utils.Precondition.check(!isSigningPoolRegistrationAsOwner, TxErrors.OUTPUT_WITH_PATH);
      }

      if (output.tokenBundle) {
        _utils.Precondition.checkIsArray(output.tokenBundle, TxErrors.OUTPUT_INVALID_TOKEN_BUNDLE);

        _utils.Precondition.check(output.tokenBundle.length <= ASSET_GROUPS_MAX);

        var _iterator7 = _createForOfIteratorHelper(output.tokenBundle),
            _step7;

        try {
          for (_iterator7.s(); !(_step7 = _iterator7.n()).done;) {
            var assetGroup = _step7.value;

            _utils.Precondition.checkIsHexString(assetGroup.policyIdHex, TxErrors.OUTPUT_INVALID_TOKEN_POLICY);

            _utils.Precondition.check(assetGroup.policyIdHex.length === TOKEN_POLICY_LENGTH * 2, TxErrors.OUTPUT_INVALID_TOKEN_POLICY);

            _utils.Precondition.checkIsArray(assetGroup.tokens);

            _utils.Precondition.check(assetGroup.tokens.length <= TOKENS_IN_GROUP_MAX);

            var _iterator8 = _createForOfIteratorHelper(assetGroup.tokens),
                _step8;

            try {
              for (_iterator8.s(); !(_step8 = _iterator8.n()).done;) {
                var token = _step8.value;

                _utils.Precondition.checkIsHexString(token.assetNameHex, TxErrors.OUTPUT_INVALID_ASSET_NAME);

                _utils.Precondition.check(token.assetNameHex.length <= TOKEN_NAME_LENGTH * 2, TxErrors.OUTPUT_INVALID_ASSET_NAME);

                _utils.Precondition.checkIsUint64Str(token.amountStr);
              }
            } catch (err) {
              _iterator8.e(err);
            } finally {
              _iterator8.f();
            }
          }
        } catch (err) {
          _iterator7.e(err);
        } finally {
          _iterator7.f();
        }
      }
    } // fee

  } catch (err) {
    _iterator5.e(err);
  } finally {
    _iterator5.f();
  }

  _utils.Precondition.checkIsValidAdaAmount(feeStr, TxErrors.FEE_INVALID); //  ttl


  if (ttlStr != null) {
    _utils.Precondition.checkIsPositiveUint64Str(ttlStr, TxErrors.TTL_INVALID);
  } // certificates


  validateCertificates(certificates); // withdrawals

  _utils.Precondition.checkIsArray(withdrawals, TxErrors.WITHDRAWALS_NOT_ARRAY);

  if (isSigningPoolRegistrationAsOwner && withdrawals.length > 0) {
    throw new Error(TxErrors.WITHDRAWALS_FORBIDDEN);
  }

  var _iterator6 = _createForOfIteratorHelper(withdrawals),
      _step6;

  try {
    for (_iterator6.s(); !(_step6 = _iterator6.n()).done;) {
      var withdrawal = _step6.value;

      _utils.Precondition.checkIsValidAdaAmount(withdrawal.amountStr);

      _utils.Precondition.checkIsValidPath(withdrawal.path);
    } // metadata could be null

  } catch (err) {
    _iterator6.e(err);
  } finally {
    _iterator6.f();
  }

  if (metadataHashHex != null) {
    _utils.Precondition.checkIsHexString(metadataHashHex, TxErrors.METADATA_INVALID);

    _utils.Precondition.check(metadataHashHex.length == 32 * 2, TxErrors.METADATA_INVALID);
  } //  validity interval start


  if (validityIntervalStartStr != null) {
    _utils.Precondition.checkIsPositiveUint64Str(validityIntervalStartStr, TxErrors.VALIDITY_INTERVAL_START_INVALID);
  }
}

function serializeAddressParams(addressTypeNibble, networkIdOrProtocolMagic, spendingPath) {
  var stakingPath = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
  var stakingKeyHashHex = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : null;
  var stakingBlockchainPointer = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : null;

  _utils.Precondition.checkIsUint8(addressTypeNibble << 4, TxErrors.OUTPUT_INVALID_ADDRESS_TYPE_NIBBLE);

  var addressTypeNibbleBuf = _utils["default"].uint8_to_buf(addressTypeNibble);

  var networkIdOrProtocolMagicBuf;

  if (addressTypeNibble === AddressTypeNibbles.BYRON) {
    _utils.Precondition.checkIsUint32(networkIdOrProtocolMagic, TxErrors.INVALID_PROTOCOL_MAGIC);

    networkIdOrProtocolMagicBuf = _utils["default"].uint32_to_buf(networkIdOrProtocolMagic);
  } else {
    _utils.Precondition.checkIsUint8(networkIdOrProtocolMagic, TxErrors.INVALID_NETWORK_ID);

    networkIdOrProtocolMagicBuf = _utils["default"].uint8_to_buf(networkIdOrProtocolMagic);
  }

  _utils.Precondition.checkIsValidPath(spendingPath, TxErrors.OUTPUT_INVALID_SPENDING_PATH);

  var spendingPathBuf = _utils["default"].path_to_buf(spendingPath);

  var stakingChoices = {
    NO_STAKING: 0x11,
    STAKING_KEY_PATH: 0x22,
    STAKING_KEY_HASH: 0x33,
    BLOCKCHAIN_POINTER: 0x44
  };
  // serialize staking info
  var stakingChoice;
  var stakingInfoBuf;

  if (!stakingPath && !stakingKeyHashHex && !stakingBlockchainPointer) {
    stakingChoice = stakingChoices.NO_STAKING;
    stakingInfoBuf = Buffer.alloc(0);
  } else if (stakingPath && !stakingKeyHashHex && !stakingBlockchainPointer) {
    stakingChoice = stakingChoices.STAKING_KEY_PATH;

    _utils.Precondition.checkIsValidPath(stakingPath, TxErrors.OUTPUT_INVALID_STAKING_KEY_PATH);

    stakingInfoBuf = _utils["default"].path_to_buf(stakingPath);
  } else if (!stakingPath && stakingKeyHashHex && !stakingBlockchainPointer) {
    var stakingKeyHash = _utils["default"].hex_to_buf(stakingKeyHashHex);

    stakingChoice = stakingChoices.STAKING_KEY_HASH;

    _utils.Precondition.check(stakingKeyHash.length === KEY_HASH_LENGTH, TxErrors.OUTPUT_INVALID_STAKING_KEY_HASH);

    stakingInfoBuf = stakingKeyHash;
  } else if (!stakingPath && !stakingKeyHashHex && stakingBlockchainPointer) {
    stakingChoice = stakingChoices.BLOCKCHAIN_POINTER;
    stakingInfoBuf = Buffer.alloc(3 * 4); // 3 x uint32

    _utils.Precondition.checkIsUint32(stakingBlockchainPointer.blockIndex, TxErrors.OUTPUT_INVALID_BLOCKCHAIN_POINTER);

    stakingInfoBuf.writeUInt32BE(stakingBlockchainPointer.blockIndex, 0);

    _utils.Precondition.checkIsUint32(stakingBlockchainPointer.txIndex, TxErrors.OUTPUT_INVALID_BLOCKCHAIN_POINTER);

    stakingInfoBuf.writeUInt32BE(stakingBlockchainPointer.txIndex, 4);

    _utils.Precondition.checkIsUint32(stakingBlockchainPointer.certificateIndex, TxErrors.OUTPUT_INVALID_BLOCKCHAIN_POINTER);

    stakingInfoBuf.writeUInt32BE(stakingBlockchainPointer.certificateIndex, 8);
  } else {
    throw new Error(TxErrors.OUTPUT_INVALID_STAKING_INFO);
  }

  var stakingChoiceBuf = Buffer.from([stakingChoice]);
  return Buffer.concat([addressTypeNibbleBuf, networkIdOrProtocolMagicBuf, spendingPathBuf, stakingChoiceBuf, stakingInfoBuf]);
}

function serializeOutputBasicParams(output, protocolMagic, networkId) {
  _utils.Precondition.checkIsValidAdaAmount(output.amountStr);

  var outputType;
  var addressBuf;

  if (output.addressHex) {
    outputType = _Ada.TxOutputTypeCodes.SIGN_TX_OUTPUT_TYPE_ADDRESS_BYTES;

    _utils.Precondition.checkIsHexString(output.addressHex, TxErrors.OUTPUT_INVALID_ADDRESS);

    _utils.Precondition.check(output.addressHex.length <= 128 * 2, TxErrors.OUTPUT_INVALID_ADDRESS);

    addressBuf = Buffer.concat([_utils["default"].uint32_to_buf(output.addressHex.length / 2), _utils["default"].hex_to_buf(output.addressHex)]);
  } else if (output.spendingPath) {
    outputType = _Ada.TxOutputTypeCodes.SIGN_TX_OUTPUT_TYPE_ADDRESS_PARAMS;
    addressBuf = serializeAddressParams(output.addressTypeNibble, output.addressTypeNibble === AddressTypeNibbles.BYRON ? protocolMagic : networkId, output.spendingPath, output.stakingPath, output.stakingKeyHashHex, output.stakingBlockchainPointer);
  } else {
    throw new Error(TxErrors.OUTPUT_UNKNOWN_TYPE);
  }

  var numassetGroups = output.tokenBundle ? output.tokenBundle.length : 0;
  return Buffer.concat([_utils["default"].uint8_to_buf(outputType), addressBuf, _utils["default"].ada_amount_to_buf(output.amountStr), _utils["default"].uint32_to_buf(numassetGroups)]);
} // TODO remove after ledger app 2.2 is widespread


function serializeOutputBasicParamsBefore_2_2(output, protocolMagic, networkId) {
  _utils.Precondition.checkIsValidAdaAmount(output.amountStr);

  if (output.addressHex) {
    _utils.Precondition.checkIsHexString(output.addressHex, TxErrors.OUTPUT_INVALID_ADDRESS);

    _utils.Precondition.check(output.addressHex.length <= 128 * 2, TxErrors.OUTPUT_INVALID_ADDRESS);

    return Buffer.concat([_utils["default"].ada_amount_to_buf(output.amountStr), _utils["default"].uint8_to_buf(_Ada.TxOutputTypeCodes.SIGN_TX_OUTPUT_TYPE_ADDRESS_BYTES), _utils["default"].hex_to_buf(output.addressHex)]);
  } else if (output.spendingPath) {
    return Buffer.concat([_utils["default"].ada_amount_to_buf(output.amountStr), _utils["default"].uint8_to_buf(_Ada.TxOutputTypeCodes.SIGN_TX_OUTPUT_TYPE_ADDRESS_PARAMS), serializeAddressParams(output.addressTypeNibble, output.addressTypeNibble === AddressTypeNibbles.BYRON ? protocolMagic : networkId, output.spendingPath, output.stakingPath, output.stakingKeyHashHex, output.stakingBlockchainPointer)]);
  } else {
    throw new Error(TxErrors.OUTPUT_UNKNOWN_TYPE);
  }
}

function serializePoolInitialParams(params) {
  _utils.Precondition.checkIsHexString(params.poolKeyHashHex, TxErrors.CERTIFICATE_POOL_INVALID_POOL_KEY_HASH);

  _utils.Precondition.check(params.poolKeyHashHex.length === KEY_HASH_LENGTH * 2, TxErrors.CERTIFICATE_POOL_INVALID_POOL_KEY_HASH);

  _utils.Precondition.checkIsHexString(params.vrfKeyHashHex, TxErrors.CERTIFICATE_POOL_INVALID_VRF_KEY_HASH);

  _utils.Precondition.check(params.vrfKeyHashHex.length === 32 * 2, TxErrors.CERTIFICATE_POOL_INVALID_VRF_KEY_HASH);

  _utils.Precondition.checkIsValidAdaAmount(params.pledgeStr, TxErrors.CERTIFICATE_POOL_INVALID_PLEDGE);

  _utils.Precondition.checkIsValidAdaAmount(params.costStr, TxErrors.CERTIFICATE_POOL_INVALID_COST);

  var marginNumeratorStr = params.margin.numeratorStr;
  var marginDenominatorStr = params.margin.denominatorStr;

  _utils.Precondition.checkIsUint64Str(marginNumeratorStr, TxErrors.CERTIFICATE_POOL_INVALID_MARGIN);

  _utils.Precondition.checkIsValidPoolMarginDenominator(marginDenominatorStr, TxErrors.CERTIFICATE_POOL_INVALID_MARGIN_DENOMINATOR); // given both are valid uint strings, the check below is equivalent to "marginNumerator <= marginDenominator"


  _utils.Precondition.checkIsValidUintStr(marginNumeratorStr, marginDenominatorStr, TxErrors.CERTIFICATE_POOL_INVALID_MARGIN);

  _utils.Precondition.checkIsHexString(params.rewardAccountHex, TxErrors.CERTIFICATE_POOL_INVALID_REWARD_ACCOUNT);

  _utils.Precondition.check(params.rewardAccountHex.length === 29 * 2, TxErrors.CERTIFICATE_POOL_INVALID_REWARD_ACCOUNT);

  _utils.Precondition.check(params.poolOwners.length <= POOL_REGISTRATION_OWNERS_MAX, TxErrors.CERTIFICATE_POOL_OWNERS_TOO_MANY);

  _utils.Precondition.check(params.relays.length <= POOL_REGISTRATION_RELAYS_MAX, TxErrors.CERTIFICATE_POOL_RELAYS_TOO_MANY);

  return Buffer.concat([_utils["default"].hex_to_buf(params.poolKeyHashHex), _utils["default"].hex_to_buf(params.vrfKeyHashHex), _utils["default"].ada_amount_to_buf(params.pledgeStr), _utils["default"].ada_amount_to_buf(params.costStr), _utils["default"].uint64_to_buf(params.margin.numeratorStr), _utils["default"].uint64_to_buf(params.margin.denominatorStr), _utils["default"].hex_to_buf(params.rewardAccountHex), _utils["default"].uint32_to_buf(params.poolOwners.length), _utils["default"].uint32_to_buf(params.relays.length)]);
}

function serializePoolOwnerParams(params) {
  var SIGN_TX_POOL_OWNER_TYPE_PATH = 1;
  var SIGN_TX_POOL_OWNER_TYPE_KEY_HASH = 2;
  var path = params.stakingPath;
  var hashHex = params.stakingKeyHashHex;

  if (path) {
    _utils.Precondition.checkIsValidPath(path, TxErrors.CERTIFICATE_POOL_OWNER_INVALID_PATH);

    var pathBuf = _utils["default"].path_to_buf(path);

    var typeBuf = Buffer.alloc(1);
    typeBuf.writeUInt8(SIGN_TX_POOL_OWNER_TYPE_PATH);
    return Buffer.concat([typeBuf, pathBuf]);
  }

  if (hashHex) {
    _utils.Precondition.checkIsHexString(hashHex, TxErrors.CERTIFICATE_POOL_OWNER_INVALID_KEY_HASH);

    _utils.Precondition.check(hashHex.length === KEY_HASH_LENGTH * 2, TxErrors.CERTIFICATE_POOL_OWNER_INVALID_KEY_HASH);

    var hashBuf = _utils["default"].hex_to_buf(hashHex);

    var _typeBuf = Buffer.alloc(1);

    _typeBuf.writeUInt8(SIGN_TX_POOL_OWNER_TYPE_KEY_HASH);

    return Buffer.concat([_typeBuf, hashBuf]);
  }

  throw new Error(TxErrors.CERTIFICATE_POOL_OWNER_INCOMPLETE);
}

function serializePoolRelayParams(relayParams) {
  var type = relayParams.type;
  var params = relayParams.params;
  var RELAY_NO = 1;
  var RELAY_YES = 2;
  var yesBuf = Buffer.alloc(1);
  yesBuf.writeUInt8(RELAY_YES);
  var noBuf = Buffer.alloc(1);
  noBuf.writeUInt8(RELAY_NO);
  var portBuf;

  if (params.portNumber) {
    _utils.Precondition.checkIsUint32(params.portNumber, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_PORT);

    _utils.Precondition.check(params.portNumber <= 65535, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_PORT);

    portBuf = Buffer.concat([yesBuf, _utils["default"].uint16_to_buf(params.portNumber)]);
  } else {
    portBuf = noBuf;
  }

  var ipv4Buf;

  if (params.ipv4) {
    _utils.Precondition.checkIsString(params.ipv4, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_IPV4);

    var ipParts = params.ipv4.split('.');

    _utils.Precondition.check(ipParts.length === 4, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_IPV4);

    var ipBytes = Buffer.alloc(4);

    for (var i = 0; i < 4; i++) {
      var ipPart = _utils["default"].safe_parseInt(ipParts[i]);

      _utils.Precondition.checkIsUint8(ipPart, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_IPV4);

      ipBytes.writeUInt8(ipPart, i);
    }

    ipv4Buf = Buffer.concat([yesBuf, ipBytes]);
  } else {
    ipv4Buf = noBuf;
  }

  var ipv6Buf;

  if (params.ipv6) {
    _utils.Precondition.checkIsString(params.ipv6, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_IPV6);

    var ipHex = params.ipv6.split(':').join('');

    _utils.Precondition.checkIsHexString(ipHex, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_IPV6);

    _utils.Precondition.check(ipHex.length === 16 * 2, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_IPV6);

    ipv6Buf = Buffer.concat([yesBuf, (0, _utils2.hex_to_buf)(ipHex)]);
  } else {
    ipv6Buf = noBuf;
  }

  var dnsBuf;

  if (params.dnsName) {
    _utils.Precondition.checkIsString(params.dnsName);

    _utils.Precondition.check(params.dnsName.length <= 64, TxErrors.CERTIFICATE_POOL_RELAY_INVALID_DNS);

    _utils.Precondition.check(/^[\x00-\x7F]*$/.test(params.dnsName), TxErrors.CERTIFICATE_POOL_RELAY_INVALID_DNS);

    _utils.Precondition.check(params.dnsName.split('').every(function (c) {
      return c.charCodeAt(0) >= 32 && c.charCodeAt(0) <= 126;
    }), TxErrors.CERTIFICATE_POOL_RELAY_INVALID_DNS);

    dnsBuf = Buffer.from(params.dnsName, "ascii");
  }

  _utils.Precondition.checkIsUint8(type);

  var typeBuf = Buffer.alloc(1);
  typeBuf.writeUInt8(type);

  switch (type) {
    case 0:
      return Buffer.concat([typeBuf, portBuf, ipv4Buf, ipv6Buf]);

    case 1:
      _utils.Precondition.check(dnsBuf, TxErrors.CERTIFICATE_POOL_RELAY_MISSING_DNS);

      return Buffer.concat([typeBuf, portBuf, dnsBuf]);

    case 2:
      _utils.Precondition.check(dnsBuf, TxErrors.CERTIFICATE_POOL_RELAY_MISSING_DNS);

      return Buffer.concat([typeBuf, dnsBuf]);
  }

  throw new Error(TxErrors.CERTIFICATE_POOL_RELAY_INVALID_TYPE);
}

function serializePoolMetadataParams(params) {
  var includeMetadataBuffer = Buffer.alloc(1);

  if (params != null) {
    includeMetadataBuffer.writeUInt8(SignTxIncluded.SIGN_TX_INCLUDED_YES);
  } else {
    includeMetadataBuffer.writeUInt8(SignTxIncluded.SIGN_TX_INCLUDED_NO);
    return includeMetadataBuffer;
  }

  var url = params.metadataUrl;

  _utils.Precondition.checkIsString(url, TxErrors.CERTIFICATE_POOL_METADATA_INVALID_URL);

  _utils.Precondition.check(url.length <= 64, TxErrors.CERTIFICATE_POOL_METADATA_INVALID_URL);

  _utils.Precondition.check(url.split('').every(function (c) {
    return c.charCodeAt(0) >= 32 && c.charCodeAt(0) <= 126;
  }), TxErrors.CERTIFICATE_POOL_METADATA_INVALID_URL);

  var hashHex = params.metadataHashHex;

  _utils.Precondition.checkIsHexString(hashHex, TxErrors.CERTIFICATE_POOL_METADATA_INVALID_HASH);

  _utils.Precondition.check(hashHex.length === 32 * 2, TxErrors.CERTIFICATE_POOL_METADATA_INVALID_HASH);

  return Buffer.concat([includeMetadataBuffer, _utils["default"].hex_to_buf(hashHex), Buffer.from(url)]);
}

function serializeGetExtendedPublicKeyParams(path) {
  _utils.Precondition.check(path, GetKeyErrors.INVALID_PATH);

  return _utils["default"].path_to_buf(path);
}

var _default = {
  HARDENED: HARDENED,
  AddressTypeNibbles: AddressTypeNibbles,
  CertificateTypes: CertificateTypes,
  KEY_HASH_LENGTH: KEY_HASH_LENGTH,
  TX_HASH_LENGTH: TX_HASH_LENGTH,
  serializeGetExtendedPublicKeyParams: serializeGetExtendedPublicKeyParams,
  validateTransaction: validateTransaction,
  serializeAddressParams: serializeAddressParams,
  serializeOutputBasicParams: serializeOutputBasicParams,
  serializeOutputBasicParamsBefore_2_2: serializeOutputBasicParamsBefore_2_2,
  serializePoolInitialParams: serializePoolInitialParams,
  serializePoolOwnerParams: serializePoolOwnerParams,
  serializePoolRelayParams: serializePoolRelayParams,
  serializePoolMetadataParams: serializePoolMetadataParams
};
exports["default"] = _default;
//# sourceMappingURL=cardano.js.map